import logging

from tfw.config import TFWENV
from tfw.networking import Scope
from tfw.components import LogInotifyObserver

from .event_handler import EventHandler

LOG = logging.getLogger(__name__)


class LogMonitoringEventHandler(EventHandler):
    """
    Monitors the output of a supervisor process (stdout, stderr) and
    sends the results to the frontend.

    Accepts messages that have a data['command'] key specifying
    a command to be executed.

    The API of each command is documented in their respective handler.
    """
    def __init__(self, key, process_name, log_tail=0):
        super().__init__(key, scope=Scope.WEBSOCKET)
        self.process_name = process_name
        self._monitor = LogInotifyObserver(
            server_connector=self.server_connector,
            supervisor_uri=TFWENV.SUPERVISOR_HTTP_URI,
            process_name=process_name,
            log_tail=log_tail
        )
        self._monitor.start()

        self.command_handlers = {
            'process_name': self.handle_process_name,
            'log_tail':     self.handle_log_tail
        }

    def handle_event(self, message):
        try:
            data = message['data']
            self.command_handlers[data['command']](data)
        except KeyError:
            LOG.error('IGNORING MESSAGE: Invalid message received: %s', message)

    def handle_process_name(self, data):
        """
        Changes the monitored process.

        :param data: TFW message data containing 'value'
                     (name of the process to monitor)
        """
        self._monitor.process_name = data['value']

    def handle_log_tail(self, data):
        """
        Sets tail length of the log the monitor will send
        to the frontend (the monitor will send back the last
        'value' characters of the log).

        :param data: TFW message data containing 'value'
                     (new tail length)
        """
        self._monitor.log_tail = data['value']

    def cleanup(self):
        self._monitor.stop()
